'''utility functions related to the collatz sequence'''

from typing import Dict, Generator


def next_collatz(n: int) -> int:
    '''returns the next number in the collatz sequence'''
    if n < 1:
        raise ValueError
    return n // 2 if n % 2 == 0 else n * 3 + 1


def collatz_iterator(n: int) -> Generator[int, None, None]:
    '''iterates through the collatz sequence from n to 1'''
    if n < 1:
        raise ValueError
    while True:
        yield n
        if n == 1:
            break
        n = next_collatz(n)


class _collatz:
    '''internal memory for finding collatz length'''
    __known_lengths: Dict[int, int] = {1: 1}

    @staticmethod
    def get_length(n: int) -> int:
        '''returns the length of collatz sequence of n'''
        if n in _collatz.__known_lengths:
            return _collatz.__known_lengths[n]
        length = 1 + _collatz.get_length(next_collatz(n))
        _collatz.__known_lengths[n] = length
        return length


def collatz_length(n: int) -> int:
    '''returns the length of collatz sequence starting at 1'''
    if n < 1:
        raise ValueError
    return _collatz.get_length(n)



'''utility generators and functions related to the fibonacci sequence'''

from typing import Generator


def fibonacci_iterator(a: int = 1, b: int = 1) -> Generator[int, None, None]:
    '''iterates through the fibonaccie sequence using starting values'''
    while True:
        yield a
        a, b = b, a + b



'''utility generators and functions related to primes'''

from typing import Generator, List


class _smart_sieve:
    '''internal sieve for prime functions

    functions will automatically use the sieve, and it will grow automatically'''

    __sieve: List[bool] = [False, False, True, True]
    __prime_list: List[int] = [2, 3]

    @staticmethod
    def __grow_sieve() -> None:
        '''doubles the size of the sieve'''
        previous_length = len(_smart_sieve.__sieve)
        _smart_sieve.__sieve += [True] * previous_length
        for prime in _smart_sieve.__prime_list:
            if previous_length % prime == 0:
                starting_index = previous_length
            else:
                starting_index = previous_length + prime - (previous_length % prime)
            for multiple in range(starting_index, len(_smart_sieve.__sieve), prime):
                _smart_sieve.__sieve[multiple] = False

    @staticmethod
    def __calculate_next_prime() -> None:
        '''finds the next prime after the last prime in prime_list, and adds it to prime_list'''
        possible_prime = _smart_sieve.__prime_list[-1] + 1
        while True:
            if possible_prime >= len(_smart_sieve.__sieve):
                _smart_sieve.__grow_sieve()
            if _smart_sieve.__sieve[possible_prime]:
                _smart_sieve.__prime_list.append(possible_prime)
                for multiple in range(possible_prime * possible_prime, len(_smart_sieve.__sieve), possible_prime):
                    _smart_sieve.__sieve[multiple] = False
                break
            possible_prime += 1

    @staticmethod
    def is_prime(n: int) -> bool:
        '''uses the sieve to check if n is prime'''
        while n > _smart_sieve.__prime_list[-1]:
            _smart_sieve.__calculate_next_prime()
        return _smart_sieve.__sieve[n]

    @staticmethod
    def nth_prime(n: int) -> int:
        '''uses the sieve's prime_list to return the nth prime'''
        while n > len(_smart_sieve.__prime_list):
            _smart_sieve.__calculate_next_prime()
        return _smart_sieve.__prime_list[n - 1]

    @staticmethod
    def get_current_sieve_size() -> int:
        '''returns the current size of the sieve, useful for optimizations'''
        return len(_smart_sieve.__sieve)


def prime_iterator() -> Generator[int, None, None]:
    '''iterates through all primes starting at 2'''
    n = 1
    while True:
        yield _smart_sieve.nth_prime(n)
        n += 1


def is_prime(n: int) -> bool:
    '''checks if n is prime'''
    if n < 2:
        return False
    if n < _smart_sieve.get_current_sieve_size():
        return _smart_sieve.is_prime(n)
    # faster to check divisibility up to the sqrt of n
    sqrt = int(n ** 0.5)
    for prime in prime_iterator():
        if prime > sqrt:
            break
        if n % prime == 0:
            return False
    return True


def nth_prime(n: int) -> int:
    '''returns the nth_prime'''
    if n < 1:
        raise ValueError
    return _smart_sieve.nth_prime(n)



'''utility functions for operations related to factors/divisors'''

from functools import reduce
from operator import mul
from typing import Callable, Dict, List, Set
from valkyrie_util.primes import nth_prime


class _smart_factorization:
    '''internal memory of factorizations,

    ensures no recalculations of factorizations'''

    __known_factorizations: Dict[int, Dict[int, int]] = {1: {}}

    @staticmethod
    def get_prime_factors(n: int, last_prime_tested: int = 1) -> Dict[int, int]:
        '''recursively calculates factorizations'''
        if n in _smart_factorization.__known_factorizations:
            return _smart_factorization.__known_factorizations[n].copy()

        current_divisor = nth_prime(last_prime_tested)

        # add the prime to factorization if not already in
        if current_divisor not in _smart_factorization.__known_factorizations:
            _smart_factorization.__known_factorizations[current_divisor] = {current_divisor: 1}

        # can stop and conclude that n is prime if current_divisor ** 2 > n
        if current_divisor * current_divisor > n:
            _smart_factorization.__known_factorizations[n] = {n: 1}
            return {n: 1}

        if n % current_divisor == 0:
            quotient = n // current_divisor
            quotient_fact = _smart_factorization.get_prime_factors(quotient, last_prime_tested)
            fact = combine_factorizations({current_divisor: 1}, quotient_fact)
            _smart_factorization.__known_factorizations[n] = fact
            return fact
        else:
            return _smart_factorization.get_prime_factors(n, last_prime_tested + 1)


def prime_factorization(n: int) -> Dict[int, int]:
    '''returns the prime factorization of n where the n equals the product of all keys ** value'''
    if n < 1:
        raise ValueError
    elif n == 1:
        return {}
    return _smart_factorization.get_prime_factors(n)


def combine_factorizations(*facts: Dict[int, int]) -> Dict[int, int]:
    '''combines all factorizations into one, equivalent to multiplying all products of each factorization'''
    new_factorization: Dict[int, int] = {}
    for fact in facts:
        for base, exponent in fact.items():
            if base in new_factorization:
                new_factorization[base] += exponent
                if new_factorization[base] == 0:
                    del new_factorization[base]
            else:
                new_factorization[base] = exponent
    return new_factorization


def divisor_count_factorization(fact: Dict[int, int]) -> int:
    '''returns the amount of divisors from a factorization'''
    product: Callable[[int, int], int] = lambda x, y: x * y
    return reduce(product, (e + 1 for b, e in fact.items()), 1)


def divisor_count(n: int) -> int:
    '''returns the amount of divisors for n'''
    return divisor_count_factorization(prime_factorization(n))


def factorization_product(fact: Dict[int, int]) -> int:
    '''returns the product which has the prime factorization of fact'''
    return reduce(mul, (pow(b, e) for b, e in fact.items()), 1)  # type:ignore


def lcm_factorization(*facts: Dict[int, int]) -> Dict[int, int]:
    '''returns the least common multiple of all factorizations input'''
    new_factorization: Dict[int, int] = {}
    for fact in facts:
        for base, exponent in fact.items():
            if base in new_factorization:
                new_factorization[base] = max(new_factorization[base], exponent)
                if new_factorization[base] == 0:
                    del new_factorization[base]
            else:
                new_factorization[base] = exponent
    return new_factorization


class _smart_divisor_sieve:
    '''internal sieve for finding divisors'''

    __sieve: List[Set[int]] = [set(), set([1]), set([1]), set([1])]

    @staticmethod
    def __grow_sieve() -> None:
        '''doubles the size of the sieve'''
        previous_length = len(_smart_divisor_sieve.__sieve)
        for i in range(previous_length):
            _smart_divisor_sieve.__sieve.append(set([1]))
        sqrt = int(len(_smart_divisor_sieve.__sieve)**0.5)
        for divisor in range(2, sqrt + 1):
            if previous_length % divisor == 0:
                starting_index = previous_length
            else:
                starting_index = previous_length + divisor - (previous_length % divisor)
            for multiple in range(starting_index, len(_smart_divisor_sieve.__sieve), divisor):
                _smart_divisor_sieve.__sieve[multiple].add(divisor)
                _smart_divisor_sieve.__sieve[multiple].add(multiple // divisor)

    @staticmethod
    def get_divisors(n: int) -> Set[int]:
        '''returns the set of proper divisors of n, growing the sieve as necessary'''
        while n >= len(_smart_divisor_sieve.__sieve):
            _smart_divisor_sieve.__grow_sieve()
        return _smart_divisor_sieve.__sieve[n].copy()


def proper_divisors(n: int) -> Set[int]:
    '''returns the set of proper divisors of n'''
    if n < 1:
        raise ValueError
    return _smart_divisor_sieve.get_divisors(n)


class _smart_divisor_sums:
    '''internal memory of divisors sum'''
    __known_sums: Dict[int, int] = {0: 0, 1: 1}

    @staticmethod
    def get_sum(n: int) -> int:
        '''returns the sum of proper divisors of n'''
        if n not in _smart_divisor_sums.__known_sums:
            _smart_divisor_sums.__known_sums[n] = sum(proper_divisors(n))
        return _smart_divisor_sums.__known_sums[n]


def proper_divisor_sum(n: int) -> int:
    '''returns the sum of all proper divisors of n'''
    if n < 1:
        raise ValueError
    return _smart_divisor_sums.get_sum(n)



'''utility generators and functions related to triangle numbers'''

from typing import Generator


def triangle_num_iterator(n: int = 1) -> Generator[int, None, None]:
    '''iterates through the triangle numbers, starting at the nth one'''
    if n < 0:
        raise ValueError
    triangle_number = nth_triangle_number(n)
    while True:
        yield triangle_number
        n += 1
        triangle_number += n


def nth_triangle_number(n: int) -> int:
    '''returns the nth triangle number'''
    if n < 0:
        raise ValueError
    return n * (n + 1) // 2



'''operations for integers larger than typical data sizes'''

from typing import List


def __normalize(large_num: List[int]) -> List[int]:
    '''normalizes the large number by making empty lists [0] and removing leading zeros'''
    if large_num == []:
        return [0]
    normalized_num = large_num.copy()
    while len(normalized_num) > 1 and normalized_num[-1] == 0:
        normalized_num.pop()
    return normalized_num


def string_to_large_number(digit_string: str) -> List[int]:
    '''converts the string of digits to a large number format'''
    return __normalize([int(c) for c in digit_string[::-1]])


def large_number_to_string(large_num: List[int]) -> str:
    '''converts the large number to a string'''
    return str().join(str(n) for n in __normalize(large_num)[::-1])


def __carry_values(large_num: List[int], base: int = 10) -> List[int]:
    '''ensures that all values in the large number are less than the base by carrying the remainder to next value'''
    result, magnitude, remainder = [], 0, 0
    while remainder > 0 or magnitude < len(large_num):
        if magnitude < len(large_num):
            remainder += large_num[magnitude]
        result.append(remainder % base)
        remainder = remainder // base
        magnitude += 1
    return result


def large_number_sum(*large_nums: List[int]) -> List[int]:
    '''sums all large numbers in the input'''
    large_sum: List[int] = []
    for addend in large_nums:
        for magnitude in range(min(len(large_sum), len(addend))):
            large_sum[magnitude] += addend[magnitude]
        if len(addend) > len(large_sum):
            large_sum += addend[len(large_sum):]
    return __normalize(__carry_values(large_sum))


def large_number_multiply(large_num: List[int], multiplier: int) -> List[int]:
    '''returns large_num * multiplier in large number format'''
    if multiplier < 0:
        raise ValueError
    large_product = [n * multiplier for n in large_num]
    return __normalize(__carry_values(large_product))



'''formulas for fast calculations'''


def range_sum_int(start: int, end: int, divisor: int = 1) -> int:
    '''returns the sum of all integers in the range [start, end] that are divisible by the divisor'''
    calc_start = (min(start, end) + divisor - 1) // divisor
    calc_end = max(start, end) // divisor
    return ((calc_start + calc_end) * (calc_end - calc_start + 1) // 2) * divisor


def sum_of_squares(n: int) -> int:
    '''returns the sum of squares for the numbers 1 to n'''
    if n < 0:
        raise ValueError
    return n * (n + 1) * (2 * n + 1) // 6



# type: ignore

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='valkyrie_util',
    version='2.3.1',
    author='David Castner',
    author_email='davidjcastner@gmail.com',
    description='A python library of various math and performance tools',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/davidjcastner/valkyrie-util',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    python_requires='>=3.8',
)
